package NsPuzzle;


import edu.uci.ics.pattis.ics23.collections.*;
import java.util.Comparator;
import stateSpace.*;



public class NsPuzzle implements Problem {

  private int N;
  private State start;
  private State stop;
  private static int up = 0, down = 1, left = 2, right = 3;
  private static String[] decode = {"up","down","left","right"};


  public NsPuzzle (int N, int[][] start, int[][] stop) throws IllegalArgumentException {
    //Check preconditions: >=1 jug, each jug size >= 1
    if (N <= 1)
      throw new IllegalStateException
      ("NsPuzzle(Problem): Illegal N ("+N+") <= 1");
    //Should check  whether dimensions of start/stop are each == N

    //Convert to internal representation of states
    this.N     = N;
    this.start = new State(start);
    this.stop  = new State(stop);
  }



    public Comparator<Problem.State> getMetric() {
    return new Comparator<Problem.State>() {
      public int compare(Problem.State pso1, Problem.State pso2) {
        return ((NsPuzzle.State)pso2).howClose(stop) - ((NsPuzzle.State)pso1).howClose(stop);
      }
    };

 }



  public class State implements Problem.State {
    int[][] puzzleState;

    private State (int[][] puzzleState) {
      //Check preconditions
      if (puzzleState.length != N)
        throw new IllegalArgumentException
        ("NsPuzzle(State): Illegal # of rows ("+puzzleState.length+") != " + N);
      for (int i=0; i<N; i++)
        if (puzzleState[i].length != N)
          throw new IllegalArgumentException
          ("NsPuzzle(State): Illegal # of columns at row["+i+"] ("+puzzleState[i]+") != " + N);
      //Should check for values 0..N^2     

      //Legal state; copy it
      this.puzzleState = new int[N][];
      for (int i = 0; i < N; i++) {
        this.puzzleState[i] = new int[N];
        for (int j = 0; j < N; j++)
          this.puzzleState[i][j] = puzzleState[i][j];
      } 
    }


    public Problem.State copy() {
      return new State(puzzleState);
    }


   //Not required by interface, but used for best-first search
   //Here are three metrics; uncomment one to see how it works
   //Hint: try the middle one with ApplicationNoFillNoEmpty
   //      try the first  one with Application
   public int howClose(State s) {
      int answer = 0;
      for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++)
          answer += (s.puzzleState[i][j] == puzzleState[i][j] ? 0 : 1);
      } 
      return answer;
    }


    public String toString() {
      String answer = "Puzzle[\n";
      for (int i = 0; i < N; i++) {
        for (int j = 0; j < N; j++)
          answer += (this.puzzleState[i][j] < 10 ? "  " : " ") + this.puzzleState[i][j];
        answer += "\n";
      } 
      return answer + "]";
    }


    public boolean equals(Object o) {
      if (o == null)
        return false;
      if (this == o)
        return true;
      if (!(o instanceof State))
        return false;

      State os = (State) o;
      if (os.puzzleState.length != N)
        return false;
      for (int i = 0; i < N; i++)
        for (int j = 0; j < N; j++)
          if (os.puzzleState[i][j] != puzzleState[i][j])
            return false;
      return true;
    }


    public int hashCode() {
      int hash = 0;
      for (int i = 0; i < N; i++)
        for (int j = 0; j < N; j++)
          hash += 31*hash + puzzleState[i][j];
      return hash;
    }

  }



  class Slide implements Problem.Operator {
    private int puzzleState;
    private int direction;

    Slide (int direction) {
      if (direction < 0 || direction > 3)
        throw new IllegalArgumentException("NsPuzzle(Slide): Illegal direction ("+direction+") not 0..4");
      this.direction = direction;
    }


    public Problem.State apply (Problem.State s) throws IllegalStateException {
      int eR=-1,eC=-1;
      State currentState = (State)s;
      State resultState  = (State)(s.copy());
      for (int i=0; i<N; i++)
        for (int j=0; j<N; j++)
          if (currentState.puzzleState[i][j] == 0) {
            eR = i;
            eC = j;
            break;
          }
        
      if (direction == up) {
        if (eR == N-1)
          throw new IllegalStateException(); 
        resultState.puzzleState[eR]  [eC] = currentState.puzzleState[eR+1][eC];
        resultState.puzzleState[eR+1][eC] = 0;
      } else if (direction == down) {
        if (eR == 0)
          throw new IllegalStateException(); 
        resultState.puzzleState[eR]  [eC] = currentState.puzzleState[eR-1][eC];
        resultState.puzzleState[eR-1][eC] = 0;
     } else if (direction == left) {
        if (eC == N-1)
          throw new IllegalStateException(); 
        resultState.puzzleState[eR][eC]   = currentState.puzzleState[eR][eC+1];
        resultState.puzzleState[eR][eC+1] = 0;
      } else if (direction == right) {
        if (eC == 0)
          throw new IllegalStateException(); 
        resultState.puzzleState[eR][eC]   = currentState.puzzleState[eR][eC-1];
        resultState.puzzleState[eR][eC-1] = 0;
      } 

      return resultState;
    }


    public String toString() {
      return "Slide " + decode[direction];
    }
  }


  public Problem.State getStartState () {
    return start;
  }

  public Problem.State getStopState  () {
    return stop;
  }

  public Set<NsPuzzle.Operator> getAllOperators() {
    Set<NsPuzzle.Operator> operators = new ArraySet<NsPuzzle.Operator>();

    operators.add(new Slide(up));
    operators.add(new Slide(down));
    operators.add(new Slide(left));
    operators.add(new Slide(right));

    return operators;
  }

}
