package stateSpace;


import java.util.Comparator;
import edu.uci.ics.pattis.ics23.collections.*;


public class Solver {

  public static String illustrateSolution (Problem.State state, List<Problem.Operator> operators) {
    if (operators == null)
      return "No solution found";

    StringBuffer solution = new StringBuffer
                                 ("Solution length = " + operators.size() + "\n" + state	+ ": Starting State\n");
    try {
      for (Problem.Operator op : operators) {
        state = op.apply(state);
        solution.append(state + ": " + op + "\n");
      }
      solution.append(state + ": Final State");
    } catch (IllegalStateException e) {
      solution.append(e.getMessage());
    }
    return solution.toString();
  }

  
  //Use breadth-first searching to find an "optimal" solution (the one
  //  with the minimum number of operators needed to transform the initial
  //  state to the final state.
  public static List<Problem.Operator> breadthFirstSolution (Problem problem) {
    return solveIt(problem, new ArrayQueue<Problem.State>());
  }

  
  //Use the closenessMetric to try to find a solution faster (look at fewer
  //  internal states), but possibly not an optimal solution.
  public static List<Problem.Operator> bestFirstSolution (Problem problem, Comparator<Problem.State> closenessMetric) {
    return solveIt(problem, new ArrayPriorityQueue<Problem.State>(closenessMetric));
  }


  //The type of exploring is an OrderedCollection, because sometimes its argument is a
  //  Queue (for breadthFirstSolution), and sometimes it is a PriorityQueue (for bestFirstSolution)
  public static List<Problem.Operator> solveIt (Problem problem, OrderedCollection<Problem.State> exploring) { 
    int operatorCount   = 0;
    Problem.State start = problem.getStartState();
    Problem.State stop  = problem.getStopState();
    Set<Problem.Operator> operators = problem.getAllOperators();

    Map<Problem.State,List<Problem.Operator>> solutions = new ArrayMap<Problem.State,List<Problem.Operator>>();          //state -> List[Operator]

    exploring.add(start);
    solutions.put(start,new ArrayList<Problem.Operator>());   //initial state -> no operations

    while(!exploring.isEmpty()) {
      Problem.State          currentState    = exploring.remove();           //state to explore next
      List<Problem.Operator> currentSolution = solutions.get(currentState);  //operators to get to this state

      for (Problem.Operator op : operators) {
        operatorCount++;

        try {
          Problem.State newState = op.apply(currentState);
          
          if (!solutions.containsKey(newState)) {
            //extend solution to include this operator
            List<Problem.Operator> newSolution = (List<Problem.Operator>)currentSolution.shallowCopy();
            newSolution.add(op);

            //if stop state, return solution
            if (newState.equals(stop)) {
              System.out.println("Operator applications = " + operatorCount);
              return newSolution;
            }

            //update solutions/exploring
            solutions.put(newState, newSolution);
            exploring.add(newState);
          }
        } catch (IllegalStateException e) {
          //Just apply next operator if this one fails to work; i.e., skip it
        }

      } 
    }

    //Failed to find solution
    return null;
  }
}
