package waterJugs;


import edu.uci.ics.pattis.ics23.collections.*;
import java.util.Comparator;
import stateSpace.*;



public class WaterJugsNoFillNoEmpty implements Problem {

  protected int[] jugSizes;
  private   State start;
  private   State stop;


  public WaterJugsNoFillNoEmpty (int[] jugSizes, int[] start, int[] stop) throws IllegalArgumentException {
    //Check preconditions: >=1 jug, each jug size >= 1
    if (jugSizes.length <= 0)
      throw new IllegalArgumentException
      ("WaterJugsNoFillNoEmpty(Problem): Illegal jugSizes.length ("+jugSizes.length+") <= 0");
    for (int i=0; i<jugSizes.length; i++)
      if (jugSizes[i] <= 0)
        throw new IllegalArgumentException
        ("WaterJugsNoFillNoEmpty(Problem): Illegal jugSizes["+i+"] ("+jugSizes[i]+") <= 0");

    //Legal sizes; copy it
    this.jugSizes = new int[jugSizes.length];
    for (int i=0; i<jugSizes.length; i++)
      this.jugSizes[i]= jugSizes[i];

    //Convert to internal representation of states
    this.start = new State(start);
    this.stop  = new State(stop);
  }


  public Comparator<Problem.State> getMetric() {
    return new Comparator<Problem.State>() {
      public int compare(Problem.State pso1, Problem.State pso2) {
        return ((WaterJugsNoFillNoEmpty.State)pso2).howClose(stop) - ((WaterJugsNoFillNoEmpty.State)pso1).howClose(stop);
      }
    };
  }

  
  
  public class State implements Problem.State {
    int[] jugState;

    private State (int[] jugState) {
      //Check preconditions
      if (jugState.length != jugSizes.length)
        throw new IllegalArgumentException
        ("WaterJugsNoFillNoEmpty(State): Illegal state.length ("+jugState.length+") != " + jugSizes.length);
      for (int i=0; i<jugSizes.length; i++)
        if (jugState[i] < 0 || jugState[i] > jugSizes[i])
          throw new IllegalArgumentException
          ("WaterJugsNoFillNoEmpty(State): Illegal state["+i+"] ("+jugState[i]+") must be >= 0 and <= " + jugSizes[i]);

      //Legal state; copy it
      this.jugState = new int[jugState.length];
      for (int i = 0; i < jugState.length; i++)
        this.jugState[i] = jugState[i];
    }


    public Problem.State copy() {
      return new State(jugState);
    }


   //Not required by interface, but used for best-first search
   //Here are three metrics; uncomment one to see how it works
   //Hint: try the middle one with ApplicationNoFillNoEmpty
   //      try the first  one with Application
   public int howClose(State s) {
      int answer = 0;
      for (int i=0; i<s.jugState.length; i++)
          //answer += Math.abs(this.jugState[i]-s.jugState[i]);
          answer += (this.jugState[i]-s.jugState[i] == 0 ? 0 : 1);
          //answer += (this.jugState[i]-s.jugState[i]) * (this.jugState[i]-s.jugState[i]);
      return answer;
    }


    public String toString() {
      String answer = "Jugs[" + jugState[0];
      for (int i = 1; i < jugState.length; i++)
        answer += "," + jugState[i];
      return answer + "]";
    }

    
    public boolean equals(Object o) {
      if (o == null)
        return false;
      if (this == o)
        return true;
      if (!(o instanceof State))
        return false;

      State os = (State) o;
      if (os.jugState.length != jugState.length)
        return false;
      for (int i = 0; i < jugState.length; i++)
        if (os.jugState[i] != jugState[i])
          return false;
      return true;
    }

    
    public int hashCode() {
      int hash = 0;
      for (int i=0; i<jugState.length; i++)
        hash += 31*hash + jugState[i];
      return hash;
    }

  }


  
  class Transfer implements Problem.Operator {
    private int from, to;

    Transfer (int from, int to) {
      if (from<0 || from>=jugSizes.length)
        throw new IllegalArgumentException("WaterJugsNoFillNoEmpty(Transfer): Illegal from ("+from+") not a legal jug #");
      if (to<0 || to>=jugSizes.length)
        throw new IllegalArgumentException("WaterJugsNoFillNoempty(Transfer): Illegal to ("+to+") not a legal jug #");
      this.from = from;
      this.to = to;
    }

    
    public Problem.State apply (Problem.State s) {
      State currentState = (State)s;
      State resultState = (State)(s.copy());
      if (currentState.jugState[from] <= (jugSizes[to] - currentState.jugState[to])) {
        //Empty from (becomes 0) into to (incremented)
        resultState.jugState[to] += currentState.jugState[from];
        resultState.jugState[from] = 0;
      } else {
        //fill to (becomes full) from from (decremented)
        resultState.jugState[from] -= jugSizes[to] - currentState.jugState[to];
        resultState.jugState[to]   =  jugSizes[to];
      }
      return resultState;
    }

    
    public String toString() {
      return "Transfer from jug " + (from+1) + " to jug " + (to+1);
    }
  }


  public Problem.State getStartState () {
    return start;
  }

  public Problem.State getStopState  () {
    return stop;
  }

  public Set<WaterJugs.Operator> getAllOperators() {
    Set<WaterJugsNoFillNoEmpty.Operator> operators = new ArraySet<WaterJugsNoFillNoEmpty.Operator>();

    //Transfer i->j and j->i  for all i != j
    for (int i = 0; i < jugSizes.length; i++)
      for (int j = i+1; j < jugSizes.length; j++) {
        operators.add(new Transfer(i, j));
        operators.add(new Transfer(j, i));
      }
    
    return operators;
  }

}